"""
Conventions for canonical CRUD endpoints.

"""
from functools import wraps

from inflection import pluralize
from marshmallow import Schema

from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import (
    dump_response_data,
    encode_count_header,
    encode_id_header,
    load_query_string_data,
    load_request_data,
    merge_data,
    require_response_data,
)
from microcosm_flask.conventions.registry import qs, request, response
from microcosm_flask.operations import Operation
from microcosm_flask.paging import OffsetLimitPage, OffsetLimitPageSchema, identity


class CRUDConvention(Convention):

    @property
    def page_cls(self):
        return OffsetLimitPage

    @property
    def page_schema(self):
        return OffsetLimitPageSchema

    def configure_search(self, ns, definition):
        """
        Register a search endpoint.

        The definition's func should be a search function, which must:
        - accept kwargs for the query string (minimally for pagination)
        - return a tuple of (items, count) where count is the total number of items
          available (in the case of pagination)

        The definition's request_schema will be used to process query string arguments.

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        paginated_list_schema = self.page_cls.make_paginated_list_schema_class(
            ns,
            definition.response_schema,
        )()

        @self.add_route(ns.collection_path, Operation.Search, ns)
        @qs(definition.request_schema)
        @response(paginated_list_schema)
        @wraps(definition.func)
        def search(**path_data):
            page = self.page_cls.from_query_string(definition.request_schema)
            result = definition.func(**merge_data(path_data, page.to_dict(func=identity)))
            response_data, headers = page.to_paginated_list(result, ns, Operation.Search)
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                paginated_list_schema,
                response_data,
                headers=headers,
                response_format=response_format,
            )

        search.__doc__ = "Search the collection of all {}".format(pluralize(ns.subject_name))

    def configure_count(self, ns, definition):
        """
        Register a count endpoint.

        The definition's func should be a count function, which must:
        - accept kwargs for the query string
        - return a count is the total number of items available

        The definition's request_schema will be used to process query string arguments.

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.add_route(ns.collection_path, Operation.Count, ns)
        @qs(definition.request_schema)
        @wraps(definition.func)
        def count(**path_data):
            request_data = load_query_string_data(definition.request_schema)
            response_data = dict()
            count = definition.func(**merge_data(path_data, request_data))
            headers = encode_count_header(count)
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                None,
                None,
                headers=headers,
                response_format=response_format,
            )

        count.__doc__ = "Count the size of the collection of all {}".format(pluralize(ns.subject_name))

    def configure_create(self, ns, definition):
        """
        Register a create endpoint.

        The definition's func should be a create function, which must:
        - accept kwargs for the request and path data
        - return a new item

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.add_route(ns.collection_path, Operation.Create, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        @wraps(definition.func)
        def create(**path_data):
            request_data = load_request_data(definition.request_schema)
            response_data = definition.func(**merge_data(path_data, request_data))
            headers = encode_id_header(response_data)
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                definition.response_schema,
                response_data,
                status_code=Operation.Create.value.default_code,
                headers=headers,
                response_format=response_format,
            )

        create.__doc__ = "Create a new {}".format(ns.subject_name)

    def configure_updatebatch(self, ns, definition):
        """
        Register an update batch endpoint.

        The definition's func should be an update function, which must:
        - accept kwargs for the request and path data
        - return a new item

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        operation = Operation.UpdateBatch

        @self.add_route(ns.collection_path, operation, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        @wraps(definition.func)
        def update_batch(**path_data):
            headers = dict()
            request_data = load_request_data(definition.request_schema)
            response_data = definition.func(**merge_data(path_data, request_data))
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                definition.response_schema,
                response_data,
                status_code=operation.value.default_code,
                headers=headers,
                response_format=response_format,
            )

        update_batch.__doc__ = "Update a batch of {}".format(ns.subject_name)

    def configure_deletebatch(self, ns, definition):
        """
        Register a delete batch endpoint.

        The definition's func should be a delete function, which must:
        - accept kwargs for path data
        - return truthy/falsey

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        operation = Operation.DeleteBatch
        request_schema = definition.request_schema or Schema()

        @self.add_route(ns.collection_path, operation, ns)
        @qs(request_schema)
        @wraps(definition.func)
        def delete_batch(**path_data):
            headers = dict()
            request_data = load_query_string_data(request_schema)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                response_schema="",
                response_data=None,
                status_code=operation.value.default_code,
                headers=headers,
                response_format=response_format,
            )

        delete_batch.__doc__ = "Delete a batch of {}".format(ns.subject_name)

    def configure_retrieve(self, ns, definition):
        """
        Register a retrieve endpoint.

        The definition's func should be a retrieve function, which must:
        - accept kwargs for path data
        - return an item or falsey

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        request_schema = definition.request_schema or Schema()

        @self.add_route(ns.instance_path, Operation.Retrieve, ns)
        @qs(request_schema)
        @response(definition.response_schema)
        @wraps(definition.func)
        def retrieve(**path_data):
            headers = dict()
            request_data = load_query_string_data(request_schema)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                definition.response_schema,
                response_data,
                headers=headers,
                response_format=response_format,
            )

        retrieve.__doc__ = "Retrieve a {} by id".format(ns.subject_name)

    def configure_delete(self, ns, definition):
        """
        Register a delete endpoint.

        The definition's func should be a delete function, which must:
        - accept kwargs for path data
        - return truthy/falsey

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        request_schema = definition.request_schema or Schema()

        @self.add_route(ns.instance_path, Operation.Delete, ns)
        @qs(request_schema)
        @wraps(definition.func)
        def delete(**path_data):
            headers = dict()
            request_data = load_query_string_data(request_schema)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                "",
                None,
                status_code=Operation.Delete.value.default_code,
                headers=headers,
                response_format=response_format,
            )

        delete.__doc__ = "Delete a {} by id".format(ns.subject_name)

    def configure_replace(self, ns, definition):
        """
        Register a replace endpoint.

        The definition's func should be a replace function, which must:
        - accept kwargs for the request and path data
        - return the replaced item

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.add_route(ns.instance_path, Operation.Replace, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        @wraps(definition.func)
        def replace(**path_data):
            headers = dict()
            request_data = load_request_data(definition.request_schema)
            # Replace/put should create a resource if not already present, but we do not
            # enforce these semantics at the HTTP layer. If `func` returns falsey, we
            # will raise a 404.
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                definition.response_schema,
                response_data,
                headers=headers,
                response_format=response_format,
            )

        replace.__doc__ = "Create or update a {} by id".format(ns.subject_name)

    def configure_update(self, ns, definition):
        """
        Register an update endpoint.

        The definition's func should be an update function, which must:
        - accept kwargs for the request and path data
        - return an updated item

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        @self.add_route(ns.instance_path, Operation.Update, ns)
        @request(definition.request_schema)
        @response(definition.response_schema)
        @wraps(definition.func)
        def update(**path_data):
            headers = dict()
            request_data = load_request_data(definition.request_schema)
            response_data = require_response_data(definition.func(**merge_data(path_data, request_data)))
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                definition.response_schema,
                response_data,
                headers=headers,
                response_format=response_format,
            )

        update.__doc__ = "Update some or all of a {} by id".format(ns.subject_name)

    def configure_createcollection(self, ns, definition):
        """
        Register create collection endpoint.

        :param ns: the namespace
        :param definition: the endpoint definition
        """
        paginated_list_schema = self.page_cls.make_paginated_list_schema_class(
            ns,
            definition.response_schema,
        )()

        @self.add_route(ns.collection_path, Operation.CreateCollection, ns)
        @request(definition.request_schema)
        @response(paginated_list_schema)
        @wraps(definition.func)
        def create_collection(**path_data):
            request_data = load_request_data(definition.request_schema)
            page = self.page_cls.from_query_string(self.page_schema(), {})

            result = definition.func(**merge_data(
                path_data,
                merge_data(
                    request_data,
                    page.to_dict(func=identity),
                ),
            ))

            response_data, headers = page.to_paginated_list(result, ns, Operation.CreateCollection)
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                paginated_list_schema,
                response_data,
                headers=headers,
                response_format=response_format,
            )

        create_collection.__doc__ = "Create the collection of {}".format(pluralize(ns.subject_name))


def configure_crud(graph, ns, mappings):
    """
    Register CRUD endpoints for a resource object.

    :param mappings: a dictionary from operations to tuple, where each tuple contains
                     the target function and zero or more marshmallow schemas according
                     to the signature of the "register_<foo>_endpoint" functions

    Example mapping:

        {
            Operation.Create: (create_foo, NewFooSchema(), FooSchema()),
            Operation.Delete: (delete_foo,),
            Operation.Retrieve: (retrieve_foo, FooSchema()),
            Operation.Search: (search_foo, SearchFooSchema(), FooSchema(), [ResponseFormats.CSV]),
        }

    """
    convention = CRUDConvention(graph)
    convention.configure(ns, mappings)



"""
Generates a Swagger definition for registered endpoints.

Note that:
 -  Swagger operations and type names use different conventions from the internal definitions
    because we want to make usage friendly for code generation (e.g. bravado)

 -  Marshmallow to JSON Schema conversion is somewhat simplistic. There are several projects
    that already implement this conversion that we could try adapting. At the moment, the
    overhead of adapting another library's conventions is too high.

 -  All resource definitions are assumed to be shared and are declared in the "definitions"
    section of the result.

 -  All errors are defined generically.


"""
from enum import Enum, unique
from logging import getLogger
from typing import (
    Dict,
    List,
    Optional,
    Set,
    Tuple,
)
from urllib.parse import unquote

from flask.globals import _request_ctx_stack
from openapi import model as swagger
from werkzeug.routing import BuildError, Rule

from microcosm_flask.conventions.registry import (
    get_qs_schema,
    get_request_schema,
    get_response_schema,
)
from microcosm_flask.errors import ErrorContextSchema, ErrorSchema, SubErrorSchema
from microcosm_flask.namespaces import Namespace
from microcosm_flask.naming import name_for
from microcosm_flask.operations import Operation
from microcosm_flask.swagger.api import build_parameter, iter_schemas
from microcosm_flask.swagger.naming import operation_name, type_name


logger = getLogger("microcosm_flask.swagger")
# Placeholder integer id used to build URIs in werkzeug before replacing with id param name.
# Use a value that is unlikely to be present somewhere else in the URI.
PLACEHOLDER_INTEGER_ID = 1111


@unique
class RequestSide(Enum):
    REQUEST = "REQUEST"
    RESPONSE = "RESPONSE"
    BOTH = "BOTH"


def build_swagger(graph, ns, operations, validate_schema=False):
    """
    Build out the top-level swagger definition.

    """
    base_path = graph.build_route_path(ns.path, ns.prefix)
    schema = swagger.Swagger(
        swagger="2.0",
        info=swagger.Info(
            title=graph.metadata.name,
            version=ns.version,
        ),
        consumes=swagger.MediaTypeList([
            swagger.MimeType("application/json"),
        ]),
        produces=swagger.MediaTypeList([
            swagger.MimeType("application/json"),
        ]),
        basePath=base_path,
        paths=swagger.Paths(),
        definitions=swagger.Definitions(),
    )
    add_paths(schema.paths, base_path, operations)
    add_definitions(schema.definitions, operations)

    if validate_schema:
        try:
            schema.validate()
        except Exception:
            logger.exception("Swagger definition did not validate against swagger schema")
            raise

    return schema


def _construct_schema_request_arguments():
    """
    Create a mapping between a given endpoint/method and its possible arguments.
    Used in aid of generating paths during swagger generation

    """
    reqctx = _request_ctx_stack.top
    url_adapter = reqctx.url_adapter
    rules: List[Rule] = url_adapter.map._rules

    return {
        (rule.endpoint, method): rule.arguments
        for rule in rules
        for method in rule.methods
    }


def add_paths(paths, base_path, operations):
    """
    Add paths to swagger.

    """
    schema_request_arguments = _construct_schema_request_arguments()

    for operation, ns, rule, func in operations:
        path = build_path(operation, ns, schema_request_arguments)
        if not path.startswith(base_path):
            continue
        method = operation.value.method.lower()
        # If there is no version number or prefix, we'd expect the base path to be ""
        # However, OpenAPI requires the minimal base path to be "/"
        # This means we need branching logic for that special case
        suffix_start = 0 if len(base_path) == 1 else len(base_path)
        paths.setdefault(
            path[suffix_start:],
            swagger.PathItem(),
        )[method] = build_operation(operation, ns, rule, func)


def add_definitions(definitions, operations):
    """
    Add definitions to swagger.

    """
    for definition_schema, request_side in iter_definitions(definitions, operations):
        if definition_schema is None:
            continue
        if isinstance(definition_schema, str):
            continue

        for name, schema in iter_schemas(definition_schema, strict_enums=request_side != RequestSide.RESPONSE):
            definitions.setdefault(name, swagger.Schema(schema))


def iter_definitions(definitions, operations):
    """
    Generate definitions to be converted to swagger schema.

    """
    # general error schema per errors.py
    for error_schema_class in [ErrorSchema, ErrorContextSchema, SubErrorSchema]:
        yield error_schema_class(), RequestSide.BOTH

    # add all request and response schemas
    for operation, obj, rule, func in operations:
        yield get_request_schema(func), RequestSide.REQUEST
        yield get_response_schema(func), RequestSide.RESPONSE


def build_path_for_integer_param(ns, operation, arguments: Set):
    """
    Build an endpoint path when the parameters are integer-valued

    When building paths for swagger, parameter names are substituted for
    path parameters. For example, the output will have '/api/v1/person/{person_id}'.
    We still use werkzeug to build those paths with placeholders for params.

    That works with UUID-valued ids but not integer-valued ones, due do a difference
    between the UUIDConverter and NumberConverter's `to_url` methods
    (see https://github.com/pallets/werkzeug/blob/master/src/werkzeug/routing.py#L1315
    and https://github.com/pallets/werkzeug/blob/master/src/werkzeug/routing.py#L1234)

    Here we instead use placeholder integers to use werkzeug functions, and replace
    them afterwards.

    """
    ordered_args = list(arguments)
    args_substitution = {
        PLACEHOLDER_INTEGER_ID + index: argument
        for index, argument in enumerate(ordered_args)
    }
    uri_templates = {
        argument: f"{placeholder}"
        for placeholder, argument in args_substitution.items()
    }
    path = unquote(ns.url_for(operation, _external=False, **uri_templates))
    for placeholder_integer, argument in args_substitution.items():
        path = path.replace(str(placeholder_integer), f"{{{argument}}}")

    return path


def create_uri_templates(arguments):
    return {
        argument: f"{{{argument}}}"
        for argument in arguments
    }


def build_path(operation: Operation, ns: Namespace, schema_request_arguments: Optional[Dict[Tuple, List[str]]] = None):
    expected_arguments: List[str] = (
        schema_request_arguments.get((ns.endpoint_for(operation), operation.value.method), [])
        if schema_request_arguments
        else []
    )

    expected_uri_templates = create_uri_templates(expected_arguments)

    try:
        # flask will sometimes try to quote '{' and '}' characters
        return unquote(ns.url_for(operation, _external=False, **expected_uri_templates))
    except (BuildError, ValueError) as error:
        # NB: The arguments were misidentified in the previous step, use the ones supplied by the error
        actual_arguments = (
            error.suggested.arguments   # type: ignore
            if isinstance(error, BuildError)
            else expected_arguments
        )

        if ns.identifier_type == "int":
            return build_path_for_integer_param(ns, operation, actual_arguments)  # type: ignore
        else:
            # we are missing some URI path parameters
            uri_templates = create_uri_templates(actual_arguments)
            return unquote(ns.url_for(operation, _external=False, **uri_templates))


def body_param(schema):
    return swagger.BodyParameter(**{
        "name": "body",
        "in": "body",
        "schema": swagger.JsonReference({
            "$ref": "#/definitions/{}".format(type_name(name_for(schema))),
        }),
    })


def header_param(name, required=False, param_type="string"):
    """
    Build a header parameter definition.

    """
    return swagger.HeaderParameterSubSchema(**{
        "name": name,
        "in": "header",
        "required": required,
        "type": param_type,
    })


def query_param(name, field):
    """
    Build a query parameter definition.

    """
    parameter = build_parameter(field)
    parameter["name"] = name
    parameter["in"] = "query"
    parameter["required"] = field.required

    return swagger.QueryParameterSubSchema(**parameter)


def path_param(name, ns):
    """
    Build a path parameter definition.

    """
    if ns.identifier_type == "uuid":
        param_type = "string"
        param_format = "uuid"
    elif ns.identifier_type == "int":
        param_type = "integer"
        param_format = None
    else:
        param_type = "string"
        param_format = None

    kwargs = {
        "name": name,
        "in": "path",
        "required": True,
        "type": param_type,
    }
    if param_format:
        kwargs["format"] = param_format
    return swagger.PathParameterSubSchema(**kwargs)


def build_operation(operation, ns, rule, func):
    """
    Build an operation definition.

    """
    swagger_operation = swagger.Operation(
        operationId=operation_name(operation, ns),
        parameters=swagger.ParametersList([
        ]),
        responses=swagger.Responses(),
        tags=[ns.subject_name],
    )

    # custom header parameter
    swagger_operation.parameters.append(
        header_param("X-Response-Skip-Null")
    )

    # path parameters
    swagger_operation.parameters.extend([
        path_param(argument, ns)
        for argument in rule.arguments
    ])

    # query string parameters
    qs_schema = get_qs_schema(func)
    if qs_schema:
        swagger_operation.parameters.extend([
            query_param(name, field)
            for name, field in qs_schema.fields.items()
        ])

    # body parameter
    request_schema = get_request_schema(func)
    if request_schema:
        swagger_operation.parameters.append(
            body_param(request_schema)
        )

    # sort parameters for predictable output
    swagger_operation.parameters.sort(key=lambda parameter: parameter["name"])

    add_responses(swagger_operation, operation, ns, func)
    return swagger_operation


def add_responses(swagger_operation, operation, ns, func):
    """
    Add responses to an operation.

    """
    # default error
    swagger_operation.responses["default"] = build_response(
        description="An error occurred",
        resource=type_name(name_for(ErrorSchema())),
    )

    if getattr(func, "__doc__", None):
        description = func.__doc__.strip().splitlines()[0]
    else:
        description = "{} {}".format(operation.value.name, ns.subject_name)

    if operation in (Operation.Upload, Operation.UploadFor):
        swagger_operation.consumes = [
            "multipart/form-data"
        ]

    # resource request
    request_resource = get_request_schema(func)
    if isinstance(request_resource, str):
        if not hasattr(swagger_operation, "consumes"):
            swagger_operation.consumes = []
        swagger_operation.consumes.append(request_resource)

    # resources response
    response_resource = get_response_schema(func)
    if isinstance(response_resource, str):
        if not hasattr(swagger_operation, "produces"):
            swagger_operation.produces = []
        swagger_operation.produces.append(response_resource)
    elif not response_resource:
        response_code = (
            204
            if operation.value.default_code == 200
            else operation.value.default_code
        )
        swagger_operation.responses[str(response_code)] = build_response(
            description=description,
        )
    else:
        swagger_operation.responses[str(operation.value.default_code)] = build_response(
            description=description,
            resource=response_resource,
        )


def build_response(description, resource=None):
    """
    Build a response definition.

    """
    response = swagger.Response(
        description=description,
    )
    if resource is not None:
        response.schema = swagger.JsonReference({
            "$ref": "#/definitions/{}".format(type_name(name_for(resource))),
        })
    return response



from marshmallow import fields


"""
AsciiEncodedStringField Marshmallow field
An extension of the marshmallow fields.String that removes unicode
characters on deserialization
"""


class AsciiEncodedStringField(fields.String):
    def _deserialize(self, value, *args, **kwargs):
        if value is not None:
            value_encoded = value.encode("ascii", "ignore")
            value = value_encoded.decode()
        return super()._deserialize(value, *args, **kwargs)
